@@ -5,3 +5,6 @@ TYPE_ATTR=INTERNAL_ATTR + '_type'
 
 #relationship-related constants
 ORDER_ATTR=INTERNAL_ATTR + '_order'
+
+#auto-property index suffix
+AUTO_PROP_INDEX_VALUE='LAST_AUTO_VALUE'
