@@ -1,6 +1,7 @@
 
 from django.db import models
 from django.db.models.fields.related import add_lazy_relation
+from django.db.models.query_utils import DeferredAttribute
 
 from neo4django import Incoming, Outgoing
 from neo4django.db import DEFAULT_DB_ALIAS
@@ -20,14 +21,19 @@ class LazyModel(object):
         actors = Relationship('Actor', ...)
     """
     def __init__(self, cls, field, name, setup_reversed):
+        self.__cls = cls
+        self.__field = field
+        self.__name = name
         self.__setup_reversed = setup_reversed
         add_lazy_relation(cls, field, name, self.__setup)
+
     def __setup(self, field, target, source):
         if not issubclass(target, NodeModel):
             raise TypeError("Relationships may only extend from Nodes.")
         self.__target = target
         self.__setup_reversed(target)
     __target = None
+
     @property
     def __model(self):
         model = self.__target
@@ -35,8 +41,15 @@ class LazyModel(object):
             raise ValueError("Lazy model not initialized!")
         else:
             return model
+
+    def __getinitargs__(self):
+        return (self.__cls, self.__field, self.__name, self.__setup_reversed)
+
     def __getattr__(self, attr):
+        if attr in ('__deepcopy__',) and self.__target is None:
+            raise AttributeError
         return getattr(self.__model, attr)
+
     def __call__(self, *args, **kwargs):
         return self.__model(*args, **kwargs)
 
@@ -64,17 +77,19 @@ class RelationshipBase(type):
                 'Model': Model,
             })
 
-    def __getattr__(cls, key):
-        if hasattr(cls, 'Model'):
-            return getattr(cls.Model, key)
-        else:
-            raise AttributeError(key)
-    def __setattr__(cls, key, value):
-        if hasattr(cls, 'Model'):
-            setattr(cls.Model, key, value)
-        else:
-            raise TypeError(
-                "Cannot assign attributes to base Relationship")
+    #TODO not necessary until we have relationship models, and leads to
+    #recursion bug
+    #def __getattr__(cls, key):
+    #    if hasattr(cls, 'Model'):
+    #        return getattr(cls.Model, key)
+    #    else:
+    #        raise AttributeError(key)
+    #def __setattr__(cls, key, value):
+    #    if hasattr(cls, 'Model'):
+    #        setattr(cls.Model, key, value)
+    #    else:
+    #        raise TypeError(
+    #            "Cannot assign attributes to base Relationship")
 
 class RelationshipModel(object):
     """
@@ -113,14 +128,6 @@ class Relationship(object):
                  single=False, related_single=False, related_name=None,
                  preserve_ordering=False, metadata={}, rel_metadata={},
                 ):
-        if related_name is None:
-            if related_single:
-                related_name = '%ss'
-            else:
-                related_name = '%s_set'
-            related_name = related_name %\
-                    (target if isinstance(target, basestring)\
-                     else target.__name__.lower())
         if direction is Outgoing:
             direction = RELATIONSHIPS_OUT
         elif direction is Incoming:
@@ -129,7 +136,7 @@ class Relationship(object):
             if not isinstance(rel_type, basestring):
                 direction = rel_type.direction
             else:
-                direction = RELATIONSHIPS_ALL
+                direction = RELATIONSHIPS_OUT
         if not isinstance(rel_type, basestring):
             if rel_type.direction != direction:
                 raise ValueError("Incompatible direction!")
@@ -139,16 +146,17 @@ class Relationship(object):
         self.__single = single
         self.direction = direction
         self.__related_single = related_single
-        self.reversed_name = related_name
+        self._related_name = related_name
         self.__ordered = preserve_ordering
         self.__meta = metadata
         self.__related_meta = rel_metadata
+
     target_model = property(lambda self: self.__target)
     ordered = property(lambda self: self.__ordered)
     meta = property(lambda self: self.__meta)
 
-
     __is_reversed = False
+
     def reverse(self, target, name):
         if self.direction is RELATIONSHIPS_IN:
             direction = RELATIONSHIPS_OUT
@@ -159,10 +167,28 @@ class Relationship(object):
         relationship = Relationship(
             target, rel_type=self.__name, direction=direction,
             single=self.__related_single, related_name=name,
-            metadata=self.__related_meta)
+            metadata=self.__related_meta, preserve_ordering=self.__ordered)
         relationship.__is_reversed = True
         return relationship
 
+    def reversed_name(self, target=None):
+        if self._related_name:
+            return self._related_name
+        else:
+            return self.get_name(target, self.__single)
+
+    @staticmethod
+    def get_name(target, single=False):
+        if single:
+            suffix = '%ss'
+        else:
+            suffix = '%s_set'
+        if isinstance(target, basestring):
+            name = target.rsplit('.',1)[-1]
+        else:
+            name = target.__name__
+        return suffix % name.lower()
+
     def contribute_to_class(self, source, name):
         if not issubclass(source, NodeModel):
             raise TypeError("Relationships may only extend from Nodes.")
@@ -182,6 +208,9 @@ class Relationship(object):
                 Bound = MultipleNodes
             bound = Bound(self, source, self.__name or name, name)
         source._meta.add_field(bound)
+        if not hasattr(source._meta, '_relationships'):
+            source._meta._relationships = {}
+        source._meta._relationships[name] = bound
         setattr(source, name, bound)
         if isinstance(self.__target, basestring):
             self.__target = LazyModel(source, self, self.__target,
@@ -190,7 +219,18 @@ class Relationship(object):
         if not self.__is_reversed:
             bound._setup_reversed(target)
 
-class BoundRelationship(AttrRouter):
+    ###################
+    # SPECIAL METHODS #
+    ###################
+
+    def __getinitargs__(self):
+        return (self.__target, self.__name, self.direction, True, self.__single,
+                self.__related_single, self.reversed_name, self.__ordered,
+                self.__meta, self.__related_meta)
+
+#subclasses DeferredAttribute to avoid being set to None in
+#django.db.models.Model.__init__().
+class BoundRelationship(AttrRouter, DeferredAttribute):
     indexed = False
     rel = None
     primary_key = False
@@ -217,12 +257,20 @@ class BoundRelationship(AttrRouter):
         if not isinstance(target, LazyModel):
             self.__rel.reverse(self.__source,
                                 self.__attname).contribute_to_class(
-                target, self.reversed_name)
+                target, self.reversed_name(self.__source))
+
     attname = name = property(lambda self: self.__attname)
 
+    @property
+    def rel_type(self):
+        return self._type
+
     def get_default(self):
         return None
 
+    def contribute_to_class(self, source, name):
+        return self.__rel.contribute_to_class(source, name)
+
     def _get_val_from_obj(self, obj):
         return self.__get__(obj)
 
@@ -281,7 +329,11 @@ class BoundRelationship(AttrRouter):
     def _load_relationships(self, node):
         #Returns all neo4j relationships attached to the provided neo4j node.
         #TODO TODO we can probably trash this function with the new backend, refs 174
-        rels = node.relationships.all([self._type])
+        if self.direction is RELATIONSHIPS_OUT:
+            rel_func = node.relationships.outgoing
+        else:
+            rel_func = node.relationships.incoming
+        rels = rel_func([self._type])
         if not hasattr(self, '_relationships'):
             attrs = {}
         else:
@@ -294,19 +346,6 @@ class BoundRelationship(AttrRouter):
 
     creation_counter = property(lambda self:self.__rel.creation_counter)
 
-    def __cmp__(self, other):
-        return cmp(self.creation_counter, other.creation_counter)
-
-    def __get__(self, obj, cls=None):
-        if obj is None: return self
-        return self._get_relationship(obj, self._state_for(obj))
-
-    def __set__(self, obj, value):
-        self._set_relationship(obj, self._state_for(obj), value)
-
-    def __delete__(self, obj):
-        self._del_relationship(obj, self._state_for(obj))
-
     def _set_relationship(self, obj, state, value):
         if value is None: # assume initialization - ignore
             return # TODO: verify that obj is unsaved!
@@ -321,9 +360,11 @@ class BoundRelationship(AttrRouter):
         neo_rel_attrs = kwargs.get('attrs', {})
         neo_rel_attrs[INTERNAL_ATTR] = True
         #TODO 'using' throughout relationships.py, #175
-        other = obj._save_neo4j_node(DEFAULT_DB_ALIAS) #HACK!
+        if obj.pk is None:
+            obj.save(using=obj.using)
+        other = obj.node
         # TODO: verify that it's ok in the reverse direction?
-        if self.direction != 'in':
+        if self.direction != 'out':
             node, other = other, node
 
         node.relationships.create(self._type, other, **neo_rel_attrs)
@@ -350,14 +391,37 @@ class BoundRelationship(AttrRouter):
     def from_neo(cls, value):
         return value
 
+    ###################
+    # SPECIAL METHODS #
+    ###################
+
+    def __getinitargs__(self):
+        return (self.__rel, self.__source, self.__type, self.__attname, self.serialize)
+
+    def __cmp__(self, other):
+        return cmp(self.creation_counter, other.creation_counter)
+
+    ######################
+    # DESCRIPTOR METHODS #
+    ######################
+
+    def __get__(self, obj, cls=None):
+        if obj is None: return self
+        return self._get_relationship(obj, self._state_for(obj))
+
+    def __set__(self, obj, value):
+        self._set_relationship(obj, self._state_for(obj), value)
+
+    def __delete__(self, obj):
+        self._del_relationship(obj, self._state_for(obj))
+
 class SingleNode(BoundRelationship):
     #BoundRelationship subclass for a single node relationship without an
     #associated relationship model.
     def _get_relationship(self, obj, state):
         if self.name in state:
             changed, result = state[self.name]
-            if changed:
-                return result
+            return result
 
         if hasattr(obj, 'node'):
             this = obj.node
@@ -400,19 +464,24 @@ class SingleNode(BoundRelationship):
         if not changed: return
         rels = self._load_relationships(node)
 
-
         #delete old relationship
         #create new relationship
 
         if other is None:
             #delete old relationship if it exists
             if hasattr(rels, 'single') and rels.single:
-                rels.single.delete()
+                rels.single.delete() #TODO this deletion should be transactional w creation
             rels.single = None
         else:
             rels.single = self._create_neo_relationship(node, other)
             #other._save_neo4j_node(DEFAULT_DB_ALIAS)
 
+    def _set_cached_relationship(self, obj, other):
+        state = BoundRelationship._state_for(obj)
+        if self.name in state and state[self.name]:
+            raise ValueError("Can't set the cache on an already initialized relationship!")
+        state[self.name] = False, other
+
 class BoundRelationshipModel(BoundRelationship):
     def __init__(self, rel, cls, relname, attname, Model):
         super(BoundRelationship, self).__init__(
@@ -513,28 +582,39 @@ class MultipleRelationships(BoundRelationshipModel): # WAIT!
         pass
 
 class RelationshipInstance(models.Manager):
+    """
+    A manager that keeps state for the many side (`MultipleNodes`) of relationships.
+    """
     def __init__(self, rel, obj):
         self.__rel = rel
         self.__obj = obj
         self._added = [] # contains domain objects
-        self._removed = set() # contains relationships
+        self._removed = [] # contains relationships
+        #holds cached domain objects (that have been added or loaded by query)
+        self._cache = []
 
     ordered = property(lambda self: self.__rel.ordered)
 
     def __save__(self, node):
         #Deletes all relationships removed since last save and adds any new
         #relatonships to the database.
+
+        #TODO this should be batched
         for relationship in self._removed:
             relationship.delete()
         for obj in self._added:
-            self.__rel._create_neo_relationship(node, obj)
-        self._removed.clear()
+            new_rel = self.__rel._create_neo_relationship(node, obj)
+            self._cache.append((new_rel, obj))
+        self._removed[:] = []
         self._added[:] = []
 
-    def _neo4j_relationships(self, node):
-        for rel in self.__rel._load_relationships(node, ordered=self.ordered):
-            if rel not in self._removed:
-                yield rel
+    def _neo4j_relationships_and_models(self, node):
+        if not self._cache:
+            self._cache = [(r, self.__rel._neo4j_instance(node, r)) for r in 
+                           self.__rel._load_relationships(node, ordered=self.ordered)]
+        for tup in self._cache:
+            if tup[0] not in self._removed:
+                yield tup
 
     @property
     def _new(self):
@@ -576,7 +656,7 @@ class RelationshipInstance(models.Manager):
                 candidate_rels = rels_by_node[obj.node]\
                         if hasattr(o, 'node') else []
                 if candidate_rels:
-                    self._removed.add(candidate_rels.pop(0))
+                    self._removed.append(candidate_rels.pop(0))
                 else:
                     try:
                         self._added.remove(obj)
@@ -584,12 +664,13 @@ class RelationshipInstance(models.Manager):
                         raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self.__obj))
         else:
             for obj in objs:
-                if obj in self._added:
-                    try:
+                try:
+                    if obj in self._added:
                         self._added.remove(obj)
-                    except ValueError:
-                        raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self.__obj))
-
+                    else:
+                        self._cache.remove(obj)
+                except ValueError:
+                    raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self.__obj))
 
     def clear(self):
         all_objs = list(self.all())
@@ -612,10 +693,10 @@ class RelationshipQuerySet(object):
         self.__rel = rel
         self.__obj = obj
 
-    def __relationships(self, node):
-        for rel in self.__inst._neo4j_relationships(node):
-            if self.__keep_relationship(rel):
-                yield rel
+    def __saved_instances(self, node):
+        for rel, item in self.__inst._neo4j_relationships_and_models(node):
+            if self.__keep_relationship(rel) and self.__keep_instance(item):
+                yield item
 
     def __iter__(self):
         removed = list(self.__inst._old)
@@ -625,12 +706,8 @@ class RelationshipQuerySet(object):
         except:
             pass
         else:
-            buffered = buffer_iterator(
-                lambda rel: (rel, self.__rel._neo4j_instance(node, rel)),
-                self.__relationships(node), size=10)
-            for rel, item in buffered:
-                if rel not in removed:
-                    yield item
+            for item in self.__saved_instances(node):
+                yield item
         for item in added:
             if self.__keep_instance(item):
                 yield item
