@@ -60,15 +60,30 @@ tests=[
 	'test_rel_metadata',
 	'test_prop_metadata',
 	'test_rel_self',
-	'test_rel_string_target']
+	'test_rel_string_target',
+	'test_auto_property',
+	'test_rel_string_type',
+	'test_get_by_id',
+	'test_in_id',
+	'test_empty_array',
+	'test_typenode_transactionality',
+	'test_queryset_str',
+	'test_in_bulk',
+	'test_auto_property_indexing',
+	'test_filter_array_member',
+	'test_autoproperty_transactionality',
+	'test_abstract_rel_inheritance',
+	'test_filter_in',
+	'test_contains',
+	'test_startswith',
+	'test_datetime_auto_now_add',
+	'test_select_related',
+	'test_rel_query_direction']
 should_fail=[
 	'test_dates',
 	'test_filter_iexact',
 	'test_relationship_model',
-	'test_datetime_auto_now_add',
 	'test_nodemodel_independence',
-	'test_str_array_property_validator',
 	'test_url_array_property_validator',
 	'test_type_query',
-	'test_model_casting_validation',
-]
+	'test_model_casting_validation']
